# External Imports
import geopandas as gpd
from pandas import DataFrame
from geopandas import GeoDataFrame
from pandera import DataFrameSchema


class ObjectIterableDataFrame:
    schema: DataFrameSchema
    data: DataFrame | GeoDataFrame

    def __init__(self, data):
        self.data = self.schema.validate(data)

    def __getattr__(self, name):
        """Delegate attribute and method access to the underlying dataframe."""
        return getattr(self.data, name)

    def __iter__(self):
        """Return an iterator for the object."""
        self._index = 0
        return self

    def __next__(self):
        """Return the next item in the iterator."""
        if self._index < len(self.data):
            row = self.data.iloc[self._index]
            iter_plot = self._row_to_object(row)
            self._index += 1
            return iter_plot
        else:
            raise StopIteration

    def __getitem__(self, item):
        """Return the item at the given index."""
        result = self.data[item]
        if isinstance(result, gpd.GeoSeries):
            return self.__class__(result)
        else:
            return result

    def __setitem__(self, key, value):
        """Set the item at the given key."""
        self.data[key] = value

    def __len__(self):
        """Return the length of the object."""
        return len(self.data)

    def _row_to_object(self, *args):
        """
        Convert a row of the dataframe to an object.

        Raises
        ------
        NotImplementedError
            If the method is not implemented by the subclass.

        Notes
        -----
        This method must be implemented by the subclass.
        """
        raise NotImplementedError("_row_to_object() must be implemented by subclass")

    def dataframe_to_geodataframe(self, crs=None) -> GeoDataFrame:
        """
        Convert the object to a GeoDataFrame using columns X and Y as geometry.

        Parameters
        ----------
        crs : str, optional
            The coordinate reference system to use for the GeoDataFrame.

        Returns
        -------
        GeoDataFrame
            A GeoDataFrame representation of the object.
        """
        try:
            gdf = GeoDataFrame(
                self.data,
                geometry=gpd.points_from_xy(self.data.X, self.data.Y),
                crs=crs,
            )
        except AttributeError:
            raise AttributeError(
                "Cannot convert to GeoDataFrame. DataFrame does not contain X and Y columns."
            )

        self.data = gdf.drop(columns=["X", "Y"])

    def geodataframe_to_dataframe(self) -> DataFrame:
        """
        Convert the object to a DataFrame by dropping the geometry column.

        Returns
        -------
        DataFrame
            A DataFrame representation of the object.
        """
        try:
            # Add X and Y columns to the dataframe from the geometry
            df = DataFrame(self.data)
            df["X"] = self.data.geometry.x
            df["Y"] = self.data.geometry.y
            df = df.drop(columns=["geometry"])
        except AttributeError:
            raise AttributeError(
                "Cannot convert to DataFrame. Object does not contain geometry column."
            )

        self.data = df



import pytest
import numpy as np
import xarray as xr
from rasterio.transform import from_origin
from rasterio.enums import Resampling

from fastfuels_core.onramps.hag_pim import (
    check_same_crs,
    check_projected_crs,
    check_resolution,
    resample_raster,
    compute_cover_from_hag,
    interpolate_hag,
    interpolate_pim,
    combine_hag_pim,
    create_plots_gdf_from_resampled_pim,
)


def create_toy_raster(
    bounds=(-12, 0, 0, 8), crs="EPSG:5070", nodata_value=0, res=(1, -1), data=None
):
    x_min, x_max, y_min, y_max = bounds

    width = int((x_max - x_min) / abs(res[0]))
    height = int((y_max - y_min) / abs(res[1]))
    transform = from_origin(x_min, y_max, abs(res[0]), abs(res[1]))
    if data is None:
        data = np.ones((height, width))
    toy_raster = xr.DataArray(
        data,
        dims=["y", "x"],
        coords={
            "x": np.linspace(x_min + res[0] / 2, x_max - res[0] / 2, width),
            "y": np.linspace(y_min - res[1] / 2, y_max + res[1] / 2, height)[::-1],
        },
        name="toy_data",
    )
    toy_raster = toy_raster.rio.write_crs(crs)
    toy_raster = toy_raster.rio.write_transform(transform)
    toy_raster.rio.write_nodata(nodata_value, inplace=True)
    return toy_raster


@pytest.fixture
def hag_4326():
    data = np.array(
        [
            [0, 1, 2, 11, 12, 3],
            [4, 5, 6, 7, 8, 13],
            [9, 14, 15, 16, 17, 18],
            [0, 19, 10, 20, 21, 22],
        ]
    )
    return create_toy_raster(
        bounds=(-6, 0, 0, 4), crs="EPSG:4326", res=(1, -1), data=data
    )


@pytest.fixture
def pim_5070():
    bounds = (-12, 0, 0, 8)
    data = np.array(
        [[0, 1, 0, 1, 0, 1], [0, 2, 0, 2, 0, 2], [0, 3, 0, 3, 0, 3], [0, 4, 0, 4, 0, 4]]
    )
    return create_toy_raster(bounds=bounds, data=data, res=(2, -2))


@pytest.fixture
def hag_5070():
    data = np.arange(24).reshape((4, 6))
    return create_toy_raster(bounds=(-12.5, -0.5, 0.5, 8.5), res=(2, -2), data=data)


def test_check_same_crs_matching(hag_5070, pim_5070):
    try:
        check_same_crs(hag_5070, pim_5070)
    except ValueError:
        pytest.fail("check_same_crs raised ValueError for matching CRS.")


def test_check_same_crs_mismatching(hag_4326, hag_5070):
    with pytest.raises(ValueError, match="do not have the same CRS"):
        check_same_crs(hag_4326, hag_5070)


def test_check_projected_crs_true(pim_5070):
    try:
        check_projected_crs(pim_5070)
    except ValueError:
        pytest.fail(
            "check_projected_crs raised ValueError for raster with a projected CRS."
        )


def test_check_projected_crs_false(hag_4326):
    with pytest.raises(ValueError, match="do not have a projected CRS"):
        check_projected_crs(hag_4326)


def test_check_resolution_true(pim_5070, hag_4326):
    try:
        check_resolution(pim_5070, hag_4326, 2)
    except ValueError:
        pytest.fail(
            "check_resolution raised ValueError for rasters with correct resolutions."
        )


def test_check_resolution_false1(pim_5070, hag_4326):
    with pytest.raises(ValueError, match="is finer than the resolution of the height"):
        check_resolution(hag_4326, pim_5070, 1)


def test_check_resolution_false2(pim_5070, hag_5070):
    with pytest.raises(ValueError, match="is finer than the desired resolution"):
        check_resolution(pim_5070, hag_5070, 100)


def test_compute_cover(hag_4326):
    min_value = 10
    desired_res = 2
    cover_raster = compute_cover_from_hag(hag_4326, min_value, desired_res)
    correct_values = np.array([[0, 0.25, 0.5], [0.5, 0.75, 1]])
    assert cover_raster.rio.resolution() == (desired_res, -desired_res)
    assert cover_raster.rio.crs == hag_4326.rio.crs
    assert cover_raster.rio.nodata == hag_4326.rio.nodata
    assert np.all(cover_raster.values == correct_values)


def test_resample_raster_pim(pim_5070):
    desired_res = 1
    resampled = resample_raster(pim_5070, desired_res, Resampling.nearest)
    correct_values = np.array(
        [
            [0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1],
            [0, 0, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1],
            [0, 0, 2, 2, 0, 0, 2, 2, 0, 0, 2, 2],
            [0, 0, 2, 2, 0, 0, 2, 2, 0, 0, 2, 2],
            [0, 0, 3, 3, 0, 0, 3, 3, 0, 0, 3, 3],
            [0, 0, 3, 3, 0, 0, 3, 3, 0, 0, 3, 3],
            [0, 0, 4, 4, 0, 0, 4, 4, 0, 0, 4, 4],
            [0, 0, 4, 4, 0, 0, 4, 4, 0, 0, 4, 4],
        ]
    )
    assert resampled.rio.resolution() == (desired_res, -desired_res)
    assert resampled.rio.crs == pim_5070.rio.crs
    assert resampled.rio.nodata == pim_5070.rio.nodata
    assert np.all(resampled.values == correct_values)


def test_resample_raster_hag(hag_4326):
    desired_res = 2
    correct_values = np.array([[10, 26, 36], [42, 61, 78]])
    resampled = resample_raster(hag_4326, desired_res, Resampling.sum)
    assert resampled.rio.resolution() == (desired_res, -desired_res)
    assert resampled.rio.crs == hag_4326.rio.crs
    assert resampled.rio.nodata == hag_4326.rio.nodata
    assert np.all(resampled.values == correct_values)


def test_interpolate_pim(pim_5070):
    correct_values = np.array(
        [[1, 1, 1, 1, 1, 1], [2, 2, 2, 2, 2, 2], [3, 3, 3, 3, 3, 3], [4, 4, 4, 4, 4, 4]]
    )
    interpolated = interpolate_pim(pim_5070)
    assert interpolated.rio.resolution() == pim_5070.rio.resolution()
    assert interpolated.rio.crs == pim_5070.rio.crs
    assert interpolated.rio.nodata == pim_5070.rio.nodata
    assert np.all(interpolated.values == correct_values)


def test_interpolate_hag(hag_5070, pim_5070):
    interpolated = interpolate_hag(hag_5070, pim_5070)
    assert interpolated.rio.resolution() == pim_5070.rio.resolution()
    assert interpolated.rio.crs == pim_5070.rio.crs
    assert interpolated.rio.nodata == pim_5070.rio.nodata
    assert interpolated.coords.equals(pim_5070.coords)
    assert np.all(interpolated.values == hag_5070.values)


def test_combine_hag_pim():
    bounds = (-3, 0, 0, 2)
    pim_data = np.arange(6).reshape((2, 3))
    hag_data = np.array([[0, 0.25, 0.5], [0.5, 0.75, 1]])
    correct_25 = np.array([[0, 0, 2], [3, 4, 5]])
    correct_50 = np.array([[0, 0, 0], [0, 4, 5]])
    correct_75 = np.array([[0, 0, 0], [0, 0, 5]])
    pim = create_toy_raster(bounds=bounds, data=pim_data)
    hag = create_toy_raster(bounds=bounds, data=hag_data)
    combined_25 = combine_hag_pim(hag, pim, 0.25)
    combined_50 = combine_hag_pim(hag, pim, 0.5)
    combined_75 = combine_hag_pim(hag, pim, 0.75)
    assert combined_25.rio.resolution() == pim.rio.resolution()
    assert combined_25.rio.crs == pim.rio.crs
    assert combined_25.rio.nodata == pim.rio.nodata
    assert combined_25.coords.equals(pim.coords)
    assert np.all(combined_25.values == correct_25)
    assert np.all(combined_50.values == correct_50)
    assert np.all(combined_75.values == correct_75)


def test_create_gdf(pim_5070):
    gdf1 = create_plots_gdf_from_resampled_pim(pim_5070)
    gdf2 = create_plots_gdf_from_resampled_pim(pim_5070, "test_name")
    assert np.all(gdf1.columns == ["Y", "X", "PLOT_ID", "geometry"])
    assert np.all(gdf2.columns == ["Y", "X", "test_name", "geometry"])
    assert np.all(gdf1.X.unique() == pim_5070["x"].values)
    assert np.all(gdf1.Y.unique() == pim_5070["y"].values)
    assert np.all(gdf1.PLOT_ID.unique() == [0, 1, 2, 3, 4])



import json
from importlib.resources import files
import numpy as np

from fastfuels_core.crown_profile_models.abc import CrownProfileModel
from numpy.typing import NDArray

DATA_PATH = files("fastfuels_core.data")
with open(DATA_PATH / "spgrp_parameters.json", "r") as f:
    SPGRP_PARAMS = json.load(f)
with open(DATA_PATH / "spcd_parameters.json", "r") as f:
    SPCD_PARAMS = json.load(f)
with open(DATA_PATH / "class_parameters.json", "r") as f:
    CLASS_PARAMS = json.load(f)

# See Purves et al. (2007) Table S2 in Supporting Information
C0_R0 = 0.503
C1_R0 = 3.126
C0_R40 = 0.5
C1_R40 = 10.0
C0_B = 0.196
C1_B = 0.511


class PurvesCrownProfile(CrownProfileModel):
    """
    Purves Crown Profile Model.

    This class computes the crown profile for a given tree based on the Purves et al. (2007) model.
    It uses species-specific parameters to calculate the crown radius at different heights; parameters
    can be found in fastfuels_core/data/spcd_parameters.json.

    Parameters
    ----------
    species_code : int
        The species code of the tree.
    dbh : float
        Diameter at breast height (cm).
    height : float
        Total height of the tree (m).
    crown_ratio : float
        Ratio of the crown length to the total height of the tree.

    Attributes
    ----------
    height : float
        Total height of the tree (m).
    crown_ratio : float
        Ratio of the crown length to the total height of the tree.
    crown_base_height : float
        Height at which the crown starts (m).
    purves_max_crown_radius : float
        Maximum radius of the crown (m).
    shape_parameter : float
        Shape parameter for the crown profile.

    Methods
    -------
    get_radius_at_height(height)
        Computes the crown radius at a given height.
    get_max_radius()
        Returns the maximum crown radius.
    """

    def __init__(
        self, species_code: int, dbh: float, height: float, crown_ratio: float
    ):
        """
        Initializes the PurvesCrownProfile model with the given parameters.

        Parameters
        ----------
        species_code : int
            The species code of the tree.
        dbh : float
            Diameter at breast height (cm).
        height : float
            Total height of the tree (m).
        crown_ratio : float
            Ratio of the crown length to the total height of the tree.

        Storing all variables as a [x,1] array allows versatile calculations no matter what the data type
        The calculations do not care if there is one item (scalar) or a large array (vector)
        By storing in a 2D array ( array[x,1], we are able to perform fast calculations without conditionals or for loop

        We utilize np.atleast_2d().T to ensure scalars and one dimensional arrays are transposed into a 2D array with shape [x,1].
        This allows the outputs of functions such as get_radius_at_height() to return data in the proper format [tree, data_from_inputs]
        """

        self.height = np.atleast_2d(height).T
        self.crown_ratio = np.atleast_2d(crown_ratio).T
        self.crown_base_height = np.atleast_2d(height - height * crown_ratio).T
        self.species_code = np.atleast_2d(species_code).T
        self.dbh = np.atleast_2d(dbh).T
        self.trait_score = vectorized_trait_score_lookup(species_code)

        # Compute maximum crown radius
        self.purves_max_crown_radius = np.atleast_2d(
            self._get_purves_max_crown_radius(species_code, dbh)
        ).T

        # Compute crown shape parameter
        self.shape_parameter = np.atleast_2d(
            self._get_purves_shape_param(species_code)
        ).T

    def get_radius_at_height(self, height: float | np.ndarray) -> float | np.ndarray:
        """
        Computes the crown radius at a given height.

        We expect inputs to be either a single value (scalar)
        Or the input should be a numpy array (vector)

        The function is able to perform calculations on both types of inputs
        and will return the same type as the input

        Parameters
        ----------
        height : float or np.ndarray
            Height(s) at which to compute the crown radius (m).
            Input can be a single value (scalar) or a numpy array (vector ndarray)
        Returns
        -------
        float or np.ndarray
            Crown radius (m) at the given height(s).
            Returns a float if input height is a scalar,
            otherwise returns a numpy array if input height is a vector.

            If height is a vector [x,], and instance variable (ie crown_base_height) is a vector [y,1]
            Then output vector will be [y,x]
        """
        height = np.asarray(height)

        radius = np.where(
            height < self.crown_base_height,
            0.0,
            self.purves_max_crown_radius
            * (np.minimum((self.height - height) / self.height, 0.95) / 0.95)
            ** self.shape_parameter,
        )
        radius = np.nan_to_num(radius, nan=0.0)

        return radius if radius.size > 1 else radius.item()

    def get_max_radius(self) -> float | np.ndarray:
        """
        Returns the maximum crown radius.

        Returns
        -------
        float or np.ndarray
            Maximum crown radius (m).
        """
        return self.get_radius_at_height(self.crown_base_height)

    def _get_purves_shape_param(self, species_code: str | NDArray):
        """
        Get shape parameters for each tree for the Purves model.

        Parameters
        ----------
        trait_score : NDArray
            Trait scores for each tree.

        Returns
        -------
        shape_parameter : NDarray
            Shape parameters for each tree.
        """
        trait_score = vectorized_trait_score_lookup(species_code)

        shape_parameter = (1.0 - trait_score) * C0_B + trait_score * C1_B
        return shape_parameter

    def _get_purves_radius(
        self, z, height, crown_base, max_crown_radius, shape_parameter
    ):
        """
        Get radius at an array of z heights using the Purves crown profile model.

        Parameters
        ----------
        z : NDarray
            Array of z coordinates of float64 type.
        height : float
            Tree height in meters.
        crown_base : float
            Crown base in meters.
        max_crown_radius : float
            Maximum radius of the tree.
        shape_parameter : float
            Purves shape parameter.

        Returns
        -------
        r : NDarray
            Radius of tree evaluated at z heights.
        """

        if z < crown_base:
            return 0.0
        if z > height:
            return 0.0

        return max_crown_radius * ((height - z) / height) ** shape_parameter

    def _get_purves_max_crown_radius(
        self, species_code: str | NDArray, dbh: float | NDArray
    ):
        """
        Gets the maximum radius of a tree for the Purves crown profile model.

        Parameters
        ----------
        trait_score : NDArray
            Trait scores for each tree.
        dbh : NDArray
            Diameter at breast height for each tree in cm.

        Returns
        -------
        r_max : NDarray
            Maximum radius of each tree in meters.
        """
        trait_score = vectorized_trait_score_lookup(species_code)

        r0j = (1 - trait_score) * C0_R0 + trait_score * C1_R0
        r40j = (1 - trait_score) * C0_R40 + trait_score * C1_R40
        max_crown_radius = r0j + (r40j - r0j) * (dbh / 40.0)
        return max_crown_radius


# useful function to look up trait scores. The function is vectorized for convenience.
def _trait_score_lookup(species_code: str | NDArray):
    return SPCD_PARAMS[str(species_code)]["PURVES_TRAIT_SCORE"]


vectorized_trait_score_lookup = np.vectorize(_trait_score_lookup)



from abc import ABC, abstractmethod


class CrownProfileModel(ABC):
    """
    Abstract base class representing a tree crown profile model.
    The crown profile model is a rotational solid that can be queried at any
    height to get a radius, or crown width, at that height.

    This abstract class provides methods to get the radius of the crown at a
    given height and to get the maximum radius of the crown.
    """

    @abstractmethod
    def get_radius_at_height(self, height):
        """
        Abstract method to get the radius of the crown at a given height.
        """
        raise NotImplementedError(
            "The get_radius method must be implemented in a subclass."
        )

    @abstractmethod
    def get_max_radius(self) -> float:
        """
        Abstract method to get the maximum radius of the crown.
        """
        raise NotImplementedError(
            "The get_max_radius method must be implemented in a subclass."
        )



# Core imports
from __future__ import annotations
import json
from abc import ABC, abstractmethod
from importlib.resources import files

# Internal imports
from fastfuels_core.base import ObjectIterableDataFrame
from fastfuels_core.point_process import run_point_process
from fastfuels_core.voxelization import VoxelizedTree, voxelize_tree
from fastfuels_core.treatments import TreatmentProtocol
from fastfuels_core.crown_profile_models.abc import CrownProfileModel
from fastfuels_core.crown_profile_models.purves import PurvesCrownProfile
from fastfuels_core.crown_profile_models.beta import BetaCrownProfile

# External Imports
import numpy as np
from numpy import ndarray

from nsvb.estimators import total_foliage_dry_weight
from pandera import DataFrameSchema, Column, Check, Index


DATA_PATH = files("fastfuels_core.data")
with open(DATA_PATH / "spgrp_parameters.json", "r") as f:
    SPGRP_PARAMS = json.load(f)
with open(DATA_PATH / "spcd_parameters.json", "r") as f:
    SPCD_PARAMS = json.load(f)
with open(DATA_PATH / "class_parameters.json", "r") as f:
    CLASS_PARAMS = json.load(f)


TREE_SCHEMA_COLS = {
    "TREE_ID": Column(int),
    "SPCD": Column(
        int,
        title="Species Code",
        description="An FIA integer species code",
    ),
    "STATUSCD": Column(
        int,
        checks=Check.isin([0, 1, 2, 3]),
        title="Status Code",
        description="0 = No status, 1 = Live, 2 = Dead, 3 = Missing",
    ),
    "DIA": Column(
        float,
        checks=[
            Check.gt(0),
            Check.lt(1200),
        ],
        nullable=True,
        title="Diameter at breast height (cm)",
        description="Diameter of the tree measured at breast height (1.37 m)",
    ),
    "HT": Column(
        float,
        checks=[
            Check.gt(0),
            Check.le(116),
        ],
        nullable=True,
        title="Height (m)",
        description="Height of the tree measured from the ground",
    ),
    "CR": Column(
        float,
        checks=Check.in_range(min_value=0, max_value=1),
        nullable=True,
        title="Crown Ratio",
        description="Ratio of the crown length to the total tree height",
    ),
}


class TreeSample(ObjectIterableDataFrame):
    schema = DataFrameSchema(
        columns={
            **TREE_SCHEMA_COLS,
            **{
                "PLOT_ID": Column(int, required=False),
                "TPA": Column(
                    dtype=float,
                    nullable=True,
                    title="Trees per Area (1/m^2)",
                    description="The number of trees per unit area that the sample "
                    "tree represents based on the plot design",
                ),
            },
        },
        coerce=True,
        index=Index(int, unique=True),
    )

    def expand_to_roi(self, process, roi, **kwargs):
        """
        Expands the trees to the Region of Interest (ROI) using a specified
        point process.

        This method generates a new set of trees that fit within the ROI
        based on the characteristics of the specified point process. It
        utilizes the 'run_point_process' function from the 'point_process'
        module to create an instance of the PointProcess class, generate tree
        locations using the specified point process, and assigns these
        locations to trees.

        Parameters
        ----------
        process : str
            The type of point process to use for expanding the trees.
            Currently, only 'inhomogeneous_poisson' is supported.
        roi : GeoDataFrame
            The Region of Interest to which the trees should be expanded. It
            should be a GeoDataFrame containing the spatial boundaries of the
            region.
        **kwargs : dict
            Additional parameters to pass to the point process. See the
            documentation for the 'point_process' module for more information.

        Returns
        -------
        Trees
            A new Trees instance where the trees have been expanded to fit
            within the ROI based on the specified point process.

        Example
        -------
        # TODO: Add example
        """
        return TreePopulation(run_point_process(process, roi, self, **kwargs))


class TreePopulation(ObjectIterableDataFrame):
    schema = DataFrameSchema(
        columns={
            **TREE_SCHEMA_COLS,
            **{
                "X": Column(
                    float,
                    nullable=True,
                    required=False,
                    title="X Coordinate (m)",
                    description="X coordinate of the tree in a projected coordinate system",
                ),
                "Y": Column(
                    float,
                    nullable=True,
                    required=False,
                    title="Y Coordinate (m)",
                    description="Y coordinate of the tree in a projected coordinate system",
                ),
            },
        },
        coerce=True,
        add_missing_columns=True,
        index=Index(int, unique=True),
    )

    def _row_to_object(self, row) -> Tree:
        """
        Convert a row of the dataframe to an object.
        """
        return Tree(
            species_code=row.SPCD,
            status_code=row.STATUSCD,
            diameter=row.DIA,
            height=row.HT,
            crown_ratio=row.CR,
            x=row.X,
            y=row.Y,
        )

    def apply_treatment(
        self, treatment: TreatmentProtocol | list[TreatmentProtocol]
    ) -> TreePopulation:
        """
        Applies one or more preconfigured silvicultural treatments to the tree population.

        Parameters
        ----------
        treatment : TreatmentProtocol | list[TreatmentProtocol]
            A single treatment or a list of treatments to apply.

        Returns
        -------
        TreePopulation
            A new TreePopulation object post treatment.
        """
        if isinstance(treatment, list):
            df = self.data.copy()
            for t in treatment:
                df = t.apply(df)
            return TreePopulation(df)
        else:
            return TreePopulation(treatment.apply(self.data))


class Tree:
    """
    An object representing an individual tree. The tree has required attributes
    that represent measurements of the tree, provide estimates of the tree's
    biomass and crown profile, and the tree's location in a projected
    coordinate system.

    Attributes
    ----------
    species_code : int
        An FIA integer species code.
    status_code : int
        1 = Live, 2 = Dead, 3 = Missing
    diameter : float
        Diameter (cm) of the tree measured at breast height (1.37 m). Units: cm.
    height : float
        Height (m) of the tree measured from the ground. Units: m.
    crown_ratio : float
        Ratio (-) of the crown length to the total tree height. Units: none.
    x : float
        X coordinate of the tree in a projected coordinate system. Units: m.
    y : float
        Y coordinate of the tree in a projected coordinate system. Units: m.
    """

    species_code: int
    status_code: int
    diameter: float
    height: float
    crown_ratio: float
    x: float
    y: float

    def __init__(
        self,
        species_code: int,
        status_code: int,
        diameter: float,
        height: float,
        crown_ratio: float,
        x=0,
        y=0,
        crown_profile_model_type="purves",
        biomass_allometry_model_type="NSVB",
    ):
        # TODO: Species code needs to be valid
        self.species_code = int(species_code)

        # TODO: Status code needs to be valid
        self.status_code = status_code

        # TODO: Diameter and height need to be greater than 0
        self.diameter = diameter
        self.height = height

        # TODO: Crown ratio needs to be between 0 and 1
        self.crown_ratio = crown_ratio

        self.x = x
        self.y = y

        if crown_profile_model_type not in ["beta", "purves"]:
            raise ValueError(
                "The crown profile model must be one of the following: 'beta' or 'purves'"
            )
        self._crown_profile_model_type = crown_profile_model_type

        available_biomass_allometry_models = ["NSVB", "jenkins"]
        if biomass_allometry_model_type == "NSVB" and self.species_group == 10:
            biomass_allometry_model_type = "jenkins"
        if biomass_allometry_model_type not in available_biomass_allometry_models:
            raise ValueError(
                f"Selected biomass allometry model: {biomass_allometry_model_type}. The biomass allometry model must be one of the following: {available_biomass_allometry_models}"
            )
        self._biomass_allometry_model_type = biomass_allometry_model_type

    @property
    def crown_length(self) -> float:
        """
        Returns the length of the tree's crown.
        """
        return self.height * self.crown_ratio

    @property
    def crown_base_height(self) -> float:
        """
        Returns the height at which the live crown starts.
        """
        return self.height - self.crown_length

    @property
    def species_group(self) -> int:
        """
        Returns the species group of the tree based on the species code.
        """
        return SPCD_PARAMS[str(self.species_code)]["SPGRP"]

    @property
    def is_live(self):
        """
        Returns True if the tree is alive, False otherwise. The Tree is alive if
        the status code is 1.
        """
        return self.status_code == 1

    @property
    def crown_profile_model(self) -> CrownProfileModel:
        if self._crown_profile_model_type == "beta":
            return BetaCrownProfile(
                self.species_code, self.crown_base_height, self.crown_length
            )
        elif self._crown_profile_model_type == "purves":
            return PurvesCrownProfile(
                self.species_code, self.diameter, self.height, self.crown_ratio
            )

    @property
    def max_crown_radius(self) -> float:
        """
        Returns the maximum radius of the tree's crown.
        """
        return self.crown_profile_model.get_max_radius()

    def get_crown_radius_at_height(self, height: float | ndarray) -> float | ndarray:
        """
        Uses the crown profile model to get the radius of the crown at a given
        height in meters.
        """
        return self.crown_profile_model.get_radius_at_height(height)

    @property
    def biomass_allometry_model(self) -> BiomassAllometryModel:
        if self._biomass_allometry_model_type == "jenkins":
            return JenkinsBiomassEquations(self.species_code, self.diameter)
        elif self._biomass_allometry_model_type == "NSVB":
            return NSVBEquations(
                self.species_code,
                self.diameter,
                self.height,
            )

    @property
    def foliage_biomass(self) -> float:
        """
        Returns the estimated foliage biomass of the tree
        """
        return self.biomass_allometry_model.estimate_foliage_biomass()

    def voxelize(
        self, horizontal_resolution: float, vertical_resolution: float, **kwargs
    ) -> VoxelizedTree:
        return voxelize_tree(self, horizontal_resolution, vertical_resolution, **kwargs)

    @classmethod
    def from_row(cls, row):
        return cls(
            species_code=int(row.SPCD),
            status_code=int(row.STATUSCD),
            diameter=row.DIA,
            height=row.HT,
            crown_ratio=row.CR,
            x=row.X,
            y=row.Y,
        )


class BiomassAllometryModel(ABC):
    """
    Abstract base class representing a tree biomass allometry model.
    The biomass allometry model is used to estimate the biomass of a tree
    based on one or more independent variables, such as diameter and height.
    """

    @abstractmethod
    def estimate_foliage_biomass(self) -> float:
        """
        Abstract method to estimate the biomass of a tree based on its diameter
        and height.
        """


class NSVBEquations(BiomassAllometryModel):
    """
    This class implements the National Scale Volume and Biomass Estimators
    modeling system.

    Information on the NSVB system can be found in the following publication:
    https://www.fs.usda.gov/research/treesearch/66998
    """

    def __init__(
        self,
        species_code: int,
        diameter: float,
        height: float,
        division: str = "",
        cull: int = 0,
        decay_code: int = 0,
        actual_height: float = None,
    ):
        if not _is_valid_spcd(species_code):
            raise ValueError(f"Species code {species_code} is not valid.")

        if not _is_valid_diameter(diameter):
            raise ValueError(f"Diameter {diameter} must be greater than 0.")

        self.species_code = species_code
        self.diameter = diameter
        self.height = height
        self.division = division
        self.cull = cull
        self.decay_code = decay_code
        self.actual_height = actual_height if actual_height else height

    def estimate_foliage_biomass(self):
        """
        Foliage weight is estimated by combining foliage weight coefficients
        and the appropriate model in Table S9a of the NSVB GTR.
        """
        dia_ft = self.diameter / 2.54
        height_ft = self.height * 3.28084
        dry_foliage_lb = total_foliage_dry_weight(
            spcd=self.species_code,
            dia=dia_ft,
            ht=height_ft,
            division=self.division,
        )
        return dry_foliage_lb * 0.453592  # Convert to kg


class JenkinsBiomassEquations(BiomassAllometryModel):
    """
    This class implements the National-Scale Biomass Estimators developed by
    Jenkins et al. (2003). These equations are used to estimate above ground and
    component biomass for 10 species groups in the United States.
    """

    def __init__(self, species_code, diameter):
        if not _is_valid_spcd(species_code):
            raise ValueError(f"Species code {species_code} is not valid.")

        if not _is_valid_diameter(diameter):
            raise ValueError(f"Diameter {diameter} must be greater than 0.")

        self.species_code = species_code
        self.diameter = diameter

        self._species_group = SPCD_PARAMS[str(species_code)]["SPGRP"]
        self._is_softwood = SPCD_PARAMS[str(species_code)]["CLASS"]
        self._sapling_adjustment = SPCD_PARAMS[str(species_code)]["SAPADJ"]

    def _estimate_above_ground_biomass(self):
        """
        Uses Equation 1 and parameters in Table 4 of Jenkins et al. 2003 to
        estimate the above ground biomass of a tree based on species group and
        diameter at breast height.

        Biomass equation: bm = Exp(β_0 + (β_1 * ln(dbh))) Where bm is above
        ground biomass (kg) for trees 2.5cm dbh and larger, dbh is the
        diameter at breast height (cm), and β_0 and β_1 are parameters
        estimated from the data.

        NOTE: This method also applies a sapling adjustment factor (sapadj) for
        trees with dbh <= 12.7cm (5 inches).
        """
        # Get the parameters for the species group
        beta_0 = SPGRP_PARAMS[str(self._species_group)]["JENKINS_AGB_b0"]
        beta_1 = SPGRP_PARAMS[str(self._species_group)]["JENKINS_AGB_b1"]

        # Estimate the above ground biomass
        biomass = np.exp(beta_0 + (beta_1 * np.log(self.diameter)))

        # Apply the sapling adjustment factor
        if self.diameter <= 12.7 and self._sapling_adjustment > 0:
            biomass *= self._sapling_adjustment

        return biomass

    def estimate_foliage_biomass(self):
        """
        Uses Equation 2 and parameters in Table 6 of Jenkins et al. 2003 to
        estimate component ratio and foliage biomass of a tree based on species
        group, hardwood classification, and diameter at breast height.

        Biomass ratio equation:
        r = Exp(β_0 + (β_1 / dbh))
        Where r is the ratio of component to total aboveground biomass for
        trees 2.5cm dbh and larger, dbh is the diameter at breast height in cm,
        and β_0 and β_1 are parameters estimated from the data.

        r is multiplied by the above ground biomass to estimate the
        foliage biomass in kg.
        """
        # Get the parameters for the hardwood class
        beta_0 = CLASS_PARAMS[str(self._is_softwood)]["foliage_b0"]
        beta_1 = CLASS_PARAMS[str(self._is_softwood)]["foliage_b1"]

        # Estimate the foliage component ratio
        ratio = np.exp(beta_0 + (beta_1 / self.diameter))

        # Estimate the foliage biomass
        foliage_biomass = self._estimate_above_ground_biomass() * ratio

        return foliage_biomass


def _is_valid_spcd(spcd: int) -> bool:
    return str(int(spcd)) in SPCD_PARAMS


def _is_valid_diameter(diameter: float) -> bool:
    return diameter > 0


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    # Create subplots for different crown ratios
    crown_ratios = [0.2, 0.4, 0.6, 0.8]
    fig, axes = plt.subplots(1, len(crown_ratios), figsize=(20, 10), sharey=True)

    for ax, crown_ratio in zip(axes, crown_ratios):
        tree_beta = Tree(
            814, 1, 15.0, 5.0, crown_ratio, 0, 0, crown_profile_model_type="beta"
        )
        tree_purves = Tree(
            814, 1, 15.0, 5.0, crown_ratio, 0, 0, crown_profile_model_type="purves"
        )

        heights = np.linspace(0, tree_beta.height, 100)
        beta_p = tree_beta.get_crown_radius_at_height(heights)
        purves_p = tree_purves.get_crown_radius_at_height(heights)

        ax.plot(beta_p, heights, "-r", label="Beta Crown Profile")
        ax.plot(purves_p, heights, "-k", label="Purves Crown Profile")
        ax.set_xlabel("Crown Radius (m)")
        ax.set_title(f"Crown Ratio = {crown_ratio}")
        ax.set_aspect("equal")
        ax.legend()

    axes[0].set_ylabel("Height (m)")
    plt.gca().set_aspect("equal", adjustable="box")
    plt.show()



# internal inputs
from fastfuels_core.disc_sampling import DiscSampler
from tests.utils import make_random_tree
from tests.utils import LIST_SPCDS


# external inputs
import numpy as np
import pytest

vector_length = 100_000


@pytest.fixture(scope="module")
def list_scpds():
    return LIST_SPCDS


@pytest.fixture(scope="module")
def tree_list():
    return [make_random_tree() for _ in range(vector_length)]


@pytest.fixture(scope="module")
def id_vector(tree_list):
    return np.array([np.random.randint(1) for t in tree_list])


@pytest.fixture(scope="module")
def dbh_vector(tree_list):
    return np.array([t.diameter for t in tree_list])


@pytest.fixture(scope="module")
def height_vector(tree_list):
    return np.array([t.height for t in tree_list])


@pytest.fixture(scope="module")
def crown_ratio_vector(tree_list):
    return np.array([t.crown_ratio for t in tree_list])


@pytest.fixture(scope="module")
def spcd_vector(tree_list):
    return np.array([t.species_code for t in tree_list])


@pytest.fixture(scope="module")
def stscd_vector(tree_list):
    return np.array([t.status_code for t in tree_list])


@pytest.fixture(scope="module")
def x_vector(tree_list):
    return np.array([t.x for t in tree_list])


@pytest.fixture(scope="module")
def y_vector(tree_list):
    return np.array([t.y for t in tree_list])


class TestDiscSampling:

    def test_create_disc_sampling(
        self,
        tree_list,
        id_vector,
        dbh_vector,
        height_vector,
        crown_ratio_vector,
        spcd_vector,
        stscd_vector,
        x_vector,
        y_vector,
    ):
        import pandas as pd

        tree_population_dict = {
            "TREE_ID": id_vector,
            "SPCD": spcd_vector,
            "STATUSCD": stscd_vector,
            "DIA": dbh_vector,
            "HT": height_vector,
            "CR": crown_ratio_vector,
            "X": x_vector,
            "Y": y_vector,
        }
        tree_population_df = pd.DataFrame(tree_population_dict)

        disc_sampling_object = DiscSampler(tree_population_df)
        print(disc_sampling_object)



"""
Tests for the InhomogeneousPoissonProcess class.
"""

# Core imports
from pathlib import Path

# Internal imports
from fastfuels_core.point_process import InhomogeneousPoissonProcess

# External imports
import numpy as np
import pandas as pd
import geopandas as gpd
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from shapely.geometry import Point, Polygon

TEST_PATH = Path(__file__).parent
TEST_DATA_PATH = TEST_PATH / "data"
TEST_FIGS_PATH = TEST_PATH / "figs"
TEST_FIGS_PATH.mkdir(exist_ok=True)

# TODO: Revisit tests with additional data


class TestGenerateTreeLocations:
    def test_basic_case(self):
        roi_gdf = gpd.read_file(TEST_DATA_PATH / "polygon.geojson")
        plots_gdf = gpd.read_file(TEST_DATA_PATH / "treemap_2016_plots_data.geojson")
        trees_df = pd.read_parquet(
            TEST_DATA_PATH / "tree_collection_from_treemap_2016.parquet"
        )

        roi_gdf = roi_gdf.to_crs(roi_gdf.estimate_utm_crs())
        plots_gdf = plots_gdf.to_crs(plots_gdf.estimate_utm_crs())

        point_process = InhomogeneousPoissonProcess()
        point_process.simulate(roi_gdf, trees_df, plots_gdf)


class TestGetStructuredCoordsGrid:
    def test_play_data_10m_resolution(self):
        polygon = Polygon(
            [
                (0, 0),
                (0, 100),
                (100, 100),
                (100, 0),
            ]
        )
        roi = gpd.GeoDataFrame(geometry=[polygon])
        x, y = InhomogeneousPoissonProcess._create_structured_coords_grid(roi, 10)

        assert x.shape == (10, 10)
        assert y.shape == (10, 10)
        assert x[0, 0] == 5
        assert y[0, 0] == 95
        assert x[9, 9] == 95
        assert y[9, 9] == 5

    def test_play_data_1m_resolution(self):
        polygon = Polygon(
            [
                (0, 0),
                (0, 100),
                (100, 100),
                (100, 0),
            ]
        )
        roi = gpd.GeoDataFrame(geometry=[polygon])
        x, y = InhomogeneousPoissonProcess._create_structured_coords_grid(roi, 1)

        assert x.shape == (100, 100)
        assert y.shape == (100, 100)
        assert x[0, 0] == 0.5
        assert y[0, 0] == 99.5
        assert x[99, 99] == 99.5
        assert y[99, 99] == 0.5


class TestCalculateTreeDensityByPlot:
    def test_treemap_2016(self):
        tree_data = pd.read_parquet(
            TEST_DATA_PATH / "tree_collection_from_treemap_2016.parquet"
        )
        plot_data = gpd.read_file(TEST_DATA_PATH / "treemap_2016_plots_data.geojson")
        plot_data = plot_data.to_crs(plot_data.estimate_utm_crs())

        # Calculate the tree density for all plots
        self._test_all_plots(tree_data, plot_data)

        # Calculate the tree density for occupied plots
        self._test_occupied_plots(tree_data, plot_data)

    @staticmethod
    def _test_all_plots(tree_data, plot_data):
        """
        This test case calculates the tree density for all plots in the plots
        data. The sum of the tree density for each plot in the plots data should
        equal the sum of the TPA for each tree in the tree data that
        belongs to that plot.

        This test case accounts for ALL PLOTS, even those with 0 density.
        """
        all_plots_with_tree_density = (
            InhomogeneousPoissonProcess._calculate_per_plot_tree_density(
                plot_data, tree_data, merge_type="left"
            )
        )

        # Make sure that all plots are in the output
        assert len(all_plots_with_tree_density) == len(plot_data)

        # Sum the TPA for each tree belonging to a plot
        plot_tpa_sum = {}
        for _, tree in tree_data.iterrows():
            plot_id = tree["PLOT_ID"]
            TPA = tree["TPA"] if tree["TPA"] > 0 else 0
            plot_tpa_sum[plot_id] = plot_tpa_sum.get(plot_id, 0) + TPA

        # Compare the calculated tree density to the expected tree density
        for _, plot in all_plots_with_tree_density.iterrows():
            plot_id = plot["PLOT_ID"]
            if plot_id not in plot_tpa_sum:
                plot_tpa_sum[plot_id] = 0
            assert np.allclose(plot["TPA"], plot_tpa_sum[plot_id])

    @staticmethod
    def _test_occupied_plots(tree_data, plot_data):
        """
        This test case calculates the tree density for occupied plots (i.e.
        those containing trees) in the plots data. The sum of the tree density
        for each plot in the plots data should equal the sum of the TPA
        for each tree in the tree data that belongs to that plot.

        This test case only accounts for OCCUPIED PLOTS, i.e. those with a
        tree density > 0. Plots without trees should not be in the output.
        """
        occupied_plots_with_tree_density = (
            InhomogeneousPoissonProcess._calculate_per_plot_tree_density(
                plot_data, tree_data, merge_type="right"
            )
        )

        # Not all plots are occupied, so the output should be less than the
        # number of plots in the plots data
        assert 0 < len(occupied_plots_with_tree_density) < len(plot_data)

        # Sum the TPA for each tree belonging to a plot
        plot_tpa_sum = {}
        for _, tree in tree_data.iterrows():
            plot_id = tree["PLOT_ID"]
            TPA = tree["TPA"] if tree["TPA"] > 0 else 0
            plot_tpa_sum[plot_id] = plot_tpa_sum.get(plot_id, 0) + TPA

        # Compare the calculated tree density to the expected tree density
        for _, plot in occupied_plots_with_tree_density.iterrows():
            plot_id = plot["PLOT_ID"]
            assert np.allclose(plot["TPA"], plot_tpa_sum[plot_id])

        # Make sure that plots with 0 density are not in the output, plots with
        # tree density > 0 should be in the output
        for _, plot in plot_data.iterrows():
            plot_id = plot["PLOT_ID"]
            if plot_id not in plot_tpa_sum:
                assert plot_id not in occupied_plots_with_tree_density["PLOT_ID"].values
            else:
                assert plot_id in occupied_plots_with_tree_density["PLOT_ID"].values


class TestCalculateAllPlotsTreeDensity:
    tree_data = pd.read_parquet(
        TEST_DATA_PATH / "tree_collection_from_treemap_2016.parquet"
    )
    plots_data = gpd.read_file(TEST_DATA_PATH / "treemap_2016_plots_data.geojson")
    plots_data = plots_data.to_crs(plots_data.estimate_utm_crs())
    plot_id_col = "PLOT_ID"

    def test_all_plots(self):
        """
        This test case calculates the tree density for all plots in the plots
        data. The sum of the tree density for each plot in the plots data should
        equal the sum of the TPA for each tree in the tree data that
        belongs to that plot.

        This test case accounts for ALL PLOTS, even those with 0 density.
        """
        all_plots_with_tree_density = (
            InhomogeneousPoissonProcess._calculate_all_plots_tree_density(
                InhomogeneousPoissonProcess, self.plots_data, self.tree_data
            )
        )

        # Make sure that all plots are in the output
        assert len(all_plots_with_tree_density) == len(self.plots_data)

        # Sum the TPA for each tree belonging to a plot
        plot_tpa_sum = {}
        for _, tree in self.tree_data.iterrows():
            plot_id = tree[self.plot_id_col]
            TPA = tree["TPA"] if tree["TPA"] > 0 else 0
            plot_tpa_sum[plot_id] = plot_tpa_sum.get(plot_id, 0) + TPA

        # Compare the calculated tree density to the expected tree density
        for _, plot in all_plots_with_tree_density.iterrows():
            plot_id = plot[self.plot_id_col]
            if plot_id not in plot_tpa_sum:
                plot_tpa_sum[plot_id] = 0
            assert np.allclose(plot["TPA"], plot_tpa_sum[plot_id])


class TestCalculateOccupiedPlotsTreeDensity:
    tree_data = pd.read_parquet(
        TEST_DATA_PATH / "tree_collection_from_treemap_2016.parquet"
    )
    plots_data = gpd.read_file(TEST_DATA_PATH / "treemap_2016_plots_data.geojson")
    plots_data = plots_data.to_crs(plots_data.estimate_utm_crs())
    plot_id_col = "PLOT_ID"

    def test_occupied_plots(self):
        """
        This test case calculates the tree density for occupied plots (i.e.
        those containing trees) in the plots data. The sum of the tree density
        for each plot in the plots data should equal the sum of the TPA
        for each tree in the tree data that belongs to that plot.

        This test case only accounts for OCCUPIED PLOTS, i.e. those with a
        tree density > 0. Plots without trees should not be in the output.
        """
        occupied_plots_with_tree_density = (
            InhomogeneousPoissonProcess._calculate_occupied_plots_tree_density(
                InhomogeneousPoissonProcess, self.plots_data, self.tree_data
            )
        )

        # Not all plots are occupied, so the output should be less than the
        # number of plots in the plots data
        assert 0 < len(occupied_plots_with_tree_density) < len(self.plots_data)

        # Sum the TPA for each tree belonging to a plot
        plot_tpa_sum = {}
        for _, tree in self.tree_data.iterrows():
            plot_id = tree[self.plot_id_col]
            TPA = tree["TPA"] if tree["TPA"] > 0 else 0
            plot_tpa_sum[plot_id] = plot_tpa_sum.get(plot_id, 0) + TPA

        # Compare the calculated tree density to the expected tree density
        for _, plot in occupied_plots_with_tree_density.iterrows():
            plot_id = plot[self.plot_id_col]
            assert np.allclose(plot["TPA"], plot_tpa_sum[plot_id])

        # Make sure that plots with 0 density are not in the output, plots with
        # tree density > 0 should be in the output
        for _, plot in self.plots_data.iterrows():
            plot_id = plot[self.plot_id_col]
            if plot_id not in plot_tpa_sum:
                assert (
                    plot_id
                    not in occupied_plots_with_tree_density[self.plot_id_col].values
                )


class TestInterpolateDataToGrid:
    def test_nearest_play_data(self):
        # Create a 4x4 grid
        x_coords = np.array([0, 1, 2, 3])
        y_coords = np.array([0, 1, 2, 3])
        x_grid, y_grid = np.meshgrid(x_coords, y_coords)

        # Create a GeoDataFrame with 4 plots. Each plot is a corner of the grid
        upper_left = Point(0, 0)
        upper_right = Point(3, 0)
        lower_left = Point(0, 3)
        lower_right = Point(3, 3)
        plots = gpd.GeoDataFrame(
            geometry=[upper_left, upper_right, lower_left, lower_right]
        )
        plots["PlotID"] = [1, 2, 3, 4]

        # Use nearest neighbor to interpolate the plotIDs to the grid
        data_to_interpolate = plots["PlotID"]
        interpolated_grid = InhomogeneousPoissonProcess._interpolate_data_to_grid(
            plots, data_to_interpolate, x_grid, y_grid, method="nearest"
        )

        # Assert that each quadrant has the correct plotID
        assert interpolated_grid.shape == (4, 4)
        assert np.allclose(interpolated_grid[:2, :2], 1)
        assert np.allclose(interpolated_grid[:2, 2:], 2)
        assert np.allclose(interpolated_grid[2:, :2], 3)
        assert np.allclose(interpolated_grid[2:, 2:], 4)


class TestInterpolateTreeDensityToGrid:
    roi = gpd.read_file(TEST_DATA_PATH / "polygon.geojson")
    roi = roi.to_crs(roi.estimate_utm_crs())
    tree_data = pd.read_parquet(
        TEST_DATA_PATH / "tree_collection_from_treemap_2016.parquet"
    )
    plots_data = gpd.read_file(TEST_DATA_PATH / "treemap_2016_plots_data.geojson")
    plots_data = plots_data.to_crs(plots_data.estimate_utm_crs())

    grid_resolution = 15
    grid_cell_area = grid_resolution**2

    create_visualization = False
    show_visualization = False

    def test_interpolate_tree_density(self):
        """
        This test case calculates the tree density for all plots in the plots
        and interpolates tree density to a structured grid.
        """
        # Build the structured grid for interpolation
        grid_x, grid_y = InhomogeneousPoissonProcess._create_structured_coords_grid(
            self.roi, self.grid_resolution
        )
        process = InhomogeneousPoissonProcess()
        tree_density_grid = (
            InhomogeneousPoissonProcess._interpolate_tree_density_to_grid(
                process,
                self.tree_data,
                self.plots_data,
                grid_x,
                grid_y,
                self.grid_resolution,
            )
        )

        # Check that the output is a 2D array with the same shape as the grid
        assert len(tree_density_grid.shape) == 2
        assert tree_density_grid.shape == grid_x.shape
        assert tree_density_grid.shape == grid_y.shape

        # Do this for a list of grid cell resolutions. All density sums should
        # be close to eachother. I.e. the density sum should not change much
        # when the grid cell resolution changes.
        intensity_sums = []
        for resolution in [0.5, 1, 2, 5, 10, 15, 30]:
            (
                grid_x,
                grid_y,
            ) = InhomogeneousPoissonProcess._create_structured_coords_grid(
                self.roi, resolution
            )
            interpolated_grid = (
                InhomogeneousPoissonProcess._interpolate_tree_density_to_grid(
                    process,
                    self.tree_data,
                    self.plots_data,
                    grid_x,
                    grid_y,
                    resolution,
                )
            )
            intensity_sums.append(interpolated_grid.sum())

        # Assert that the density sums are within a few trees of eachother
        assert np.allclose(intensity_sums, intensity_sums[0], atol=4)

        if self.create_visualization or self.show_visualization:
            self.visualize_data(grid_x, grid_y, tree_density_grid)

    def visualize_data(self, grid_x, grid_y, all_plots_grid):
        fig, ax = plt.subplots(figsize=(10, 8))
        ax.imshow(
            all_plots_grid,
            extent=[grid_x.min(), grid_x.max(), grid_y.min(), grid_y.max()],
        )
        plt.xlabel("X Coordinate")
        plt.ylabel("Y Coordinate")
        plt.title("Tree Intensity Interpolated to Grid")

        if self.create_visualization:
            plt.savefig(TEST_FIGS_PATH / "tree_intensity_grid.png")
        if self.show_visualization:
            plt.show()


class TestInterpolatePlotIdToGrid:
    roi = gpd.read_file(TEST_DATA_PATH / "polygon.geojson")
    roi = roi.to_crs(roi.estimate_utm_crs())
    tree_data = pd.read_parquet(
        TEST_DATA_PATH / "tree_collection_from_treemap_2016.parquet"
    )
    plots_data = gpd.read_file(TEST_DATA_PATH / "treemap_2016_plots_data.geojson")
    plots_data = plots_data.to_crs(plots_data.estimate_utm_crs())
    plot_id_col = "PLOT_ID"

    grid_resolution = 15
    grid_cell_area = grid_resolution**2

    create_visualization = False
    show_visualization = False

    def test_interpolate_plot_id(self):
        """
        This test case calculates the tree density for occupied plots (i.e.
        those containing trees) in the plots data. The sum of the tree density
        for each plot in the plots data should equal the sum of the TPA
        for each tree in the tree data that belongs to that plot.

        This test case only accounts for OCCUPIED PLOTS, i.e. those with a
        tree density > 0. Plots without trees should not be in the output.
        """
        # Build the structured grid for interpolation
        grid_x, grid_y = InhomogeneousPoissonProcess._create_structured_coords_grid(
            self.roi, self.grid_resolution
        )
        process = InhomogeneousPoissonProcess()
        plot_id_grid = InhomogeneousPoissonProcess._interpolate_plot_id_to_grid(
            process,
            self.tree_data,
            self.plots_data,
            grid_x,
            grid_y,
        )

        # Check that the output is a 2D array with the same shape as the grid
        assert len(plot_id_grid.shape) == 2
        assert plot_id_grid.shape == grid_x.shape
        assert plot_id_grid.shape == grid_y.shape

        # Get the occupied plots from the plots data
        occupied_plots_df = (
            InhomogeneousPoissonProcess._calculate_per_plot_tree_density(
                self.plots_data, self.tree_data, merge_type="right"
            )
        )
        occupied_plots_df = occupied_plots_df.to_crs(self.roi.crs)
        plots_inside_roi = gpd.sjoin(
            occupied_plots_df, self.roi, how="inner", predicate="within"
        )
        occupied_plots_set = set(plot_id_grid.flatten().tolist())

        # Check that all plots in the output are within the ROI
        for _, plot in plots_inside_roi.iterrows():
            plot_id = plot[self.plot_id_col]
            assert plot_id in occupied_plots_set
        for plot_id in occupied_plots_set:
            assert plot_id in occupied_plots_df[self.plot_id_col].values

        if self.create_visualization or self.show_visualization:
            self.visualize_data(grid_x, grid_y, plot_id_grid, occupied_plots_df)

    def visualize_data(self, grid_x, grid_y, occupied_plots_grid, occupied_plots_df):
        # Create a consistent mapping of plot IDs to randomized indices
        plot_ids = occupied_plots_df[self.plot_id_col].unique()
        plot_indices = list(range(len(plot_ids)))
        plot_id_to_index = {
            pid: plot_indices[i] for i, pid in enumerate(sorted(plot_ids))
        }
        num_colors = len(plot_id_to_index)

        # Create a colormap
        colors = plt.cm.get_cmap("Set3", num_colors)
        cmap = mcolors.ListedColormap(colors.colors)

        # Map grid values to indices based on plot_id_to_index
        grid_color_indices = np.vectorize(plot_id_to_index.get)(occupied_plots_grid)
        grid_color_indices[np.isnan(grid_color_indices)] = -1  # Handle NaN values

        # Display the grid
        plt.figure(figsize=(10, 8))
        plt.imshow(
            grid_color_indices,
            cmap=cmap,
            extent=[grid_x.min(), grid_x.max(), grid_y.min(), grid_y.max()],
        )
        # plt.colorbar()

        # Overlay plot locations
        for plot_id, idx in plot_id_to_index.items():
            plot_color = colors.colors[idx]
            plot_subset = occupied_plots_df[
                occupied_plots_df[self.plot_id_col] == plot_id
            ]
            x = plot_subset.geometry.centroid.x
            y = plot_subset.geometry.centroid.y
            plt.scatter(
                x, y, color=plot_color, edgecolor="black", label=f"Plot {plot_id}"
            )

        plt.xlabel("X Coordinate")
        plt.ylabel("Y Coordinate")
        plt.title("PLOT_ID Interpolated to Grid")

        # Create custom legend
        plt.legend(
            handles=[
                plt.Line2D(
                    [0],
                    [0],
                    marker="o",
                    color="w",
                    label=f"Plot {pid}",
                    markerfacecolor=colors.colors[plot_id_to_index[pid]],
                    markersize=10,
                )
                for pid in plot_ids
            ]
        )

        if self.create_visualization:
            plt.savefig(TEST_FIGS_PATH / "plot_id_grid.png")
        if self.show_visualization:
            plt.show()


class TestGenerateTreeCounts:
    def test_null_case(self):
        density_grid = np.zeros((10, 10))
        tree_counts = InhomogeneousPoissonProcess._generate_tree_counts(density_grid)
        assert np.allclose(tree_counts, 0)

    def test_play_data(self):
        density_grid = np.array([0, 1, 2, 3])
        tree_counts = InhomogeneousPoissonProcess._generate_tree_counts(density_grid)
        assert tree_counts[0] == 0
        assert tree_counts[:-1].sum() > 0

    def test_seed(self):
        seed = 42
        np.random.seed(seed)
        density_grid = np.array([100, 100, 100, 100])
        tree_counts = InhomogeneousPoissonProcess._generate_tree_counts(density_grid)
        assert np.allclose(tree_counts, np.array([96, 107, 88, 103]))

    def test_treemap_2016_data(self):
        roi = gpd.read_file(TEST_DATA_PATH / "polygon.geojson")
        roi = roi.to_crs(roi.estimate_utm_crs())
        plot_data = gpd.read_file(TEST_DATA_PATH / "treemap_2016_plots_data.geojson")
        plot_data = plot_data.to_crs(plot_data.estimate_utm_crs())
        tree_data = pd.read_parquet(
            TEST_DATA_PATH / "tree_collection_from_treemap_2016.parquet"
        )
        self._test_real_data(roi, 5, tree_data, plot_data)

    def _test_real_data(self, roi, resolution, tree_data, plot_data):
        grid_x, grid_y = InhomogeneousPoissonProcess._create_structured_coords_grid(
            roi, resolution
        )
        process = InhomogeneousPoissonProcess()
        # process._set_seed(124678)
        tree_density_grid = (
            InhomogeneousPoissonProcess._interpolate_tree_density_to_grid(
                process,
                tree_data,
                plot_data,
                grid_x,
                grid_y,
                resolution,
            )
        )

        # Draw tree counts from the tree density grid
        tree_counts_sums = []
        for _ in range(100):
            tree_counts = InhomogeneousPoissonProcess._generate_tree_counts(
                tree_density_grid
            )
            tree_counts_sums.append(tree_counts.sum())

        # Assert that tree density sum is in the range of tree counts sums
        tree_counts_sums = np.array(tree_counts_sums)
        tree_density_sum = tree_density_grid.sum()
        assert tree_counts_sums.min() < tree_density_sum < tree_counts_sums.max()


class TestCalculateTreeIndices:
    """
    Tests for the _get_flattened_tree_indices method of the
    InhomogeneousPoissonProcess class.
    """

    process = InhomogeneousPoissonProcess()

    def test_empty_grid(self):
        """
        Test with an empty grid (all zeros). Should return an empty array.
        """
        count_grid = np.zeros((5, 5), dtype=int)
        expected_indices = np.array([], dtype=int)
        calculated_indices = self.process._get_flattened_tree_indices(count_grid)
        np.testing.assert_array_equal(calculated_indices, expected_indices)

    def test_uniform_grid(self):
        """
        Test with a uniform grid (same count in each cell). Should return indices
        with equal counts for each cell.
        """
        count = 3
        count_grid = np.full((4, 4), count)
        expected_indices = self._get_expected_indices(count_grid)
        calculated_indices = self.process._get_flattened_tree_indices(count_grid)
        np.testing.assert_array_equal(calculated_indices, expected_indices)

    def test_varied_grid(self):
        """
        Test with a varied count grid. Should return indices that reflect
        the distribution of tree counts.
        """
        count_grid = np.array([[1, 2], [0, 3]])
        expected_indices = self._get_expected_indices(count_grid)
        calculated_indices = self.process._get_flattened_tree_indices(count_grid)
        np.testing.assert_array_equal(calculated_indices, expected_indices)

    def test_large_grid(self):
        """
        Test with a larger grid to ensure scalability.
        """
        count_grid = np.random.randint(0, 5, (10, 10))
        expected_indices = self._get_expected_indices(count_grid)
        calculated_indices = self.process._get_flattened_tree_indices(count_grid)
        np.testing.assert_array_equal(calculated_indices, expected_indices)

    @staticmethod
    def _get_expected_indices(grid):
        """
        This function manually creates the expected indices array for a given
        count grid without the fancy numpy tricks. This is used to test the
        logic of the _get_flattened_tree_indices method.

        The way this works is that for each cell in the grid we get the count
        value at that cell. Then, there should be an entry in that list with
        the index of that cell repeated count times.
        """
        indices_list = []
        for i in range(grid.shape[0]):
            for j in range(grid.shape[1]):
                index = i * grid.shape[1] + j
                count = grid[i, j]
                indices_list.extend([index] * count)
        return np.array(indices_list)


class TestGetCellIndices:
    process = InhomogeneousPoissonProcess()

    def test_get_cell_indices_case_1(self):
        # Mock data
        count_grid = np.array([[1, 3], [2, 4]])
        tree_indices = np.array([0, 1, 1, 1, 2, 2, 3, 3, 3, 3])

        # Expected result
        expected_result = self._get_expected_result(count_grid)

        # Call the method
        result = self.process._get_grid_cell_indices(tree_indices, count_grid)

        # Assert the result
        assert np.array_equal(
            result, expected_result
        ), "The _get_cell_indices method does not return the expected result."

    def test_get_cell_indices_case_2(self):
        # Mock data
        count_grid = np.array([[2, 2], [2, 2]])
        tree_indices = np.array([0, 0, 1, 1, 2, 2, 3, 3])

        # Expected result
        expected_result = self._get_expected_result(count_grid)

        # Call the method
        result = self.process._get_grid_cell_indices(tree_indices, count_grid)

        # Assert the result
        assert np.array_equal(
            result, expected_result
        ), "The _get_cell_indices method does not return the expected result."

    @staticmethod
    def _get_expected_result(count_grid):
        """
        This function manually creates the expected result for a given
        count grid and tree indices without the fancy numpy tricks. This is
        used to test the logic of the _get_cell_indices method.
        """
        x_indices = []
        y_indices = []
        for i in range(count_grid.shape[0]):
            for j in range(count_grid.shape[1]):
                for _ in range(count_grid[i, j]):
                    x_indices.append(i)
                    y_indices.append(j)
        return np.array(x_indices), np.array(y_indices)


class TestCalculateTreeCoordinates:
    process = InhomogeneousPoissonProcess()
    process._set_seed(42)

    def test_basic_case(self):
        # Build mock data.
        # 1x2 grid of plots
        # 1 tree in plot 0, 2 trees in plot 1
        cell_i = np.array([0, 0, 0])
        cell_j = np.array([0, 1, 1])
        grid_resolution = 1
        x = np.array([0.5, 1.5])
        y = np.array([0.5, 0.5])
        grid_x, grid_y = np.meshgrid(x, y)

        # Call the method
        result = self.process._calculate_tree_coordinates(
            cell_i, cell_j, grid_resolution, grid_x, grid_y
        )

        # Make sure that the results make sense
        x_tree_coords = result[0]
        y_tree_coords = result[1]
        assert x_tree_coords.shape == y_tree_coords.shape
        assert x_tree_coords.shape == cell_i.shape
        assert y_tree_coords.shape == cell_j.shape

        # First tree should be in the range of the first plot
        assert 0 <= x_tree_coords[0] <= 1
        assert 0 <= y_tree_coords[0] <= 1

        # Second tree should be in the range of the second plot
        assert 1 <= x_tree_coords[1] <= 2
        assert 0 <= y_tree_coords[1] <= 1

        # Third tree should be in the range of the second plot
        assert 1 <= x_tree_coords[2] <= 2
        assert 0 <= y_tree_coords[2] <= 1


class TestCountTreesPerPlot:
    process = InhomogeneousPoissonProcess()
    plot_id_list = [1, 1, 2, 2, 2, 3]
    num_trees = len(plot_id_list)
    trees = pd.DataFrame({"PLOT_ID": plot_id_list})
    tree_locations = pd.DataFrame(
        {
            "X": np.random.rand(num_trees),
            "Y": np.random.rand(num_trees),
            "PLOT_ID": plot_id_list,
        }
    )

    def test_count_trees_per_plot(self):
        num_trees_per_plot = self.process._count_trees_per_plot(
            self.trees, self.tree_locations
        )
        assert num_trees_per_plot[1] == 2
        assert num_trees_per_plot[2] == 3
        assert num_trees_per_plot[3] == 1

    def test_count_trees_per_plot_empty_trees(self):
        empty_trees = pd.DataFrame(columns=["PLOT_ID"])
        num_trees_per_plot = self.process._count_trees_per_plot(self.trees, empty_trees)
        assert num_trees_per_plot[1] == 0
        assert num_trees_per_plot[2] == 0
        assert num_trees_per_plot[3] == 0


class TestSampleTreesWithinPlots:
    process = InhomogeneousPoissonProcess()
    trees = pd.DataFrame(
        {
            "PLOT_ID": [1, 1, 2, 2, 2, 3],
            "TREE_ID": [1, 2, 3, 4, 5, 6],
            "TPA": [1, 2, 3, 4, 5, 6],
        }
    )
    num_trees_per_plot = pd.Series([2, 3, 1], index=[1, 2, 3])

    def test_sample_trees_within_plots(self):
        sampled_trees = self.process._sample_trees_within_plots(
            self.trees, self.num_trees_per_plot
        )
        assert len(sampled_trees) == self.num_trees_per_plot.sum()

        # Make sure that each sampled tree belongs to the correct plot
        for _, tree in sampled_trees.iterrows():
            sampled_plot_id = tree["PLOT_ID"]
            sampled_tree_id = tree["TREE_ID"]
            original_tree = self.trees[self.trees["TREE_ID"] == sampled_tree_id]
            original_plot_id = original_tree["PLOT_ID"].values[0]
            assert sampled_plot_id == original_plot_id

    def test_sample_trees_within_plots_empty(self):
        empty_trees = pd.DataFrame(columns=["PLOT_ID", "TPA"])
        sampled_trees = self.process._sample_trees_within_plots(
            empty_trees, self.num_trees_per_plot
        )
        assert len(sampled_trees) == 0


class TestAssignTreeLocations:
    process = InhomogeneousPoissonProcess()
    sampled_trees = pd.DataFrame(
        {"PLOT_ID": [1, 1, 2, 2, 2, 3], "TPA": [1, 2, 3, 4, 5, 6]}
    )
    tree_locations = pd.DataFrame(
        {"X": np.random.rand(6), "Y": np.random.rand(6), "PLOT_ID": [1, 1, 2, 2, 2, 3]}
    )
